/*
    lm78.c - Part of a Linux module for reading sensor data.
    Copyright (c) 1997, 1998  Alexander Larsson <alla@lysator.liu.se>,
    Frodo Looijaard <frodol@dds.nl> and Philip Edelbrock <phil@netroedge.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <linux/config.h>

/* The next define disables the redefinition of kernel_version. As it is
   already defined and initialized in lm78.c, this would result in twice
   defined symbols. We must include <linux/version.h> by hand now. */
#define __NO_VERSION__

/* Hack to allow the use of CONFIG_MODVERSIONS. In the kernel itself, this
   is arranged from the main Makefile. */
#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/module.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include "lm78.h"
#include "main.h"

/* These are functions used only internally. */
static int lm78_read_value(int address);

static int lm_type = 0;
#define LMTYPE_78 1
#define LMTYPE_78_J 2
#define LMTYPE_79 3

/* This function initializes the LM78. If it is found, and initializing
   succeeds, 0 is returned; if not, an error value is returned. */
int Init_LM78(void)
{
  int res;

  if (check_region(LM78_BASE, LM78_EXTENT))
    return -ENODEV;

  outb_p( LM78_DUMMY, LM78_ADDRESS_REG );
  res = inb_p(LM78_ADDRESS_REG);
  inb_p(LM78_DATA_REG); /* reset busy flag */

  if ((res & 0x7F) != LM78_DUMMY)
    return -EIO;

  request_region(LM78_BASE, LM78_EXTENT, "lm78");

  /* Reset all except Watchdog values and last conversion values
     This sets fan-divs to 2, among others */
  LM78_WRITE(LM78_CONFIG, 0x80 );

  res = lm78_read_value(LM78_RESET);
  if (res & 0x80)
     lm_type = LMTYPE_79;
  else if (res & 0x40)
     lm_type = LMTYPE_78_J;
   else
     lm_type = LMTYPE_78;

  /* Start monitoring */
  LM78_WRITE(LM78_CONFIG, (LM78_READ(LM78_CONFIG) & 0xf7) | 0x01);

  if (lm_type == LMTYPE_78) 
    printk("LM78");
  else if (lm_type == LMTYPE_78_J)
    printk("LM78-J");
  else if (lm_type == LMTYPE_79)
    printk("LM79");
  printk(" detected and installed.\n");
#ifdef DEBUG
  printk("LM78 Serial bus address 0x%X.\n",lm78_read_value(LM78_SMBUSADDR));
#endif
  return 0;
}

/* This function cleans up after the module. Only call it after a succesful
   Init_LM78 call! */
void Cleanup_LM78(void)
{
   release_region(LM78_BASE, LM78_EXTENT);
}

/* This function tries to update the current sensor data and to put it
   in human-readable form in buf. It returns the number of characters
   written to buf. 
   TODO: Check length of buf. */
int print_lm78_values (char *buf)
{
  int len = 0;
  long v;
  int fan;

  /* Redundant test for sensor_data.valid - better safe then sorry! */
  if (sensor_data.valid) {

#ifdef LM78_VCORE
    v = sensor_data.Vcore;
    len += sprintf(buf+len, "Vcore:  %2d.%02dV\n", (int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "Vcore: not supported\n");
#endif

#ifdef LM78_VCORE2
    v = sensor_data.Vcore2;
    len += sprintf(buf+len, "Vcore2: %2d.%02dV\n", (int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "Vcore2: not supported\n");
#endif

#ifdef LM78_3_3V
    v = sensor_data.V3_3v;
    len += sprintf(buf+len, "+3.3V:  %2d.%02dV\n", (int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "+3.3V: not supported\n");
#endif

#ifdef LM78_5V
    v = sensor_data.V5v;
    len += sprintf(buf+len, "+5.0V:  %2d.%02dV\n", (int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "+5.0V: not supported\n" );
#endif

#ifdef LM78_M5V
    v = sensor_data.Vm5v;
    len += sprintf(buf+len, "-5.0V: %3d.%02dV\n", -(int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "-5.0V: not supported\n" );
#endif

#ifdef LM78_12V
    v = sensor_data.V12v;
    len += sprintf(buf+len, "+12V:   %2d.%02dV\n", (int) v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "+12V : not supported\n" );
#endif

#ifdef LM78_M12V
    v = sensor_data.Vm12v;
    len += sprintf(buf+len, "-12V:  %3d.%02dV\n", (int) -v/100, (int) v%100 );
#else
    len += sprintf(buf+len, "-12V : not supported\n" );
#endif

#ifdef LM78_VID
  v = sensor_data.vid;
  
  if ((lm_type == LMTYPE_78) || (lm_type == LMTYPE_78_J))
    v |= 0x10;
 
    if (v <= 0x01)
        len += sprintf(buf+len, "VID:  2.0%dV\n", (0x01 - v) * 5 );
    else if (v <= 0x04)
        len += sprintf(buf+len, "VID:  1.%dV\n", (0x04 - v) * 5 + 85);
    else if (v <= 0x0f)
        len += sprintf(buf+len, "VID:  reserved (unknown) value\n");
    else if (v <= 0x15)
        len += sprintf(buf+len, "VID:  3.%dV\n", (0x15 - v));
    else if (v <= 0x1e)
        len += sprintf(buf+len, "VID:  2.%dV\n", (0x1e - v) + 1);
    else 
    	len += sprintf(buf+len, "VID:  No core installed.\n"); 
#endif

    len += sprintf(buf+len, "Mainboard temperature: %d degrees\n",
                   sensor_data.temp_mb );

    fan = sensor_data.fan[0];
    len += sprintf(buf+len, "Fan 1: ");
    if (fan==255)
      len += sprintf(buf+len, "0 rpm (not present?)\n");
    else
      len += sprintf(buf+len, "%d rpm\n", LM78_RPM(fan,2));


    fan = sensor_data.fan[1];
    len += sprintf(buf+len, "Fan 2: ");
    if (fan==255)
      len += sprintf(buf+len, "0 rpm (not present?)\n");
    else
      len += sprintf(buf+len, "%d rpm\n", LM78_RPM(fan,2));

    fan = sensor_data.fan[2];
    len += sprintf(buf+len, "Fan 3: ");
    if (fan==255)
      len += sprintf(buf+len, "0 rpm (not present?)\n");
    else
      len += sprintf(buf+len, "%d rpm\n", LM78_RPM(fan,2));
  }
  return len;
}


/* Read a value from the LM78. Theoretically, several drivers can use
   the LM78 at once by using this function. In practice, this is unlikely,
   but still it won't do any harm. */
int lm78_read_value(int address) 
{
  while (LM78_BUSY) {
    current->state = TASK_INTERRUPTIBLE;
    current->timeout = jiffies + 10;
    schedule();
  }
  return LM78_READ(address);
}

/* Reread the sensor information into struct sensor_data. */
void update_lm78_values(void)
{
  /* TODO: grab Spin lock on sensor_data */


#ifdef LM78_VCORE
  sensor_data.Vcore = ((long) lm78_read_value(LM78_VCORE)) * 16 * NORMAL_VCORE / 100000;
#endif

#ifdef LM78_VCORE2
  sensor_data.Vcore2 = ((long) lm78_read_value(LM78_VCORE2)) * 16 * NORMAL_VCORE2 / 100000;
#endif

#ifdef LM78_3_3V
  sensor_data.V3_3v = ((long) lm78_read_value(LM78_3_3V)) * 16 * NORMAL_3_3V / 100000;
#endif

#ifdef LM78_5V
  sensor_data.V5v = ((long) lm78_read_value(LM78_5V)) * 16 * NORMAL_5V / 100000;
#endif

#ifdef LM78_M5V
  sensor_data.Vm5v = ((long) lm78_read_value(LM78_M5V)) * 16 * NORMAL_M5V / 100000;
#endif

#ifdef LM78_12V
  sensor_data.V12v = ((long) lm78_read_value(LM78_12V)) * 16 * NORMAL_12V / 100000;
#endif

#ifdef LM78_M12V
  sensor_data.Vm12v = ((long) lm78_read_value(LM78_M12V)) * 16 * NORMAL_M12V / 100000;
#endif

#ifdef LM78_VID
  if (! sensor_data.valid) {
    sensor_data.vid = ((long) lm78_read_value(LM78_VID)) & 0x0f;
    if(lm_type == LMTYPE_79)
      sensor_data.vid |= ((((long) lm78_read_value(LM78_RESET)) & 0x01) << 4);
  }
#endif

  sensor_data.temp_mb= lm78_read_value(LM78_TEMPERATURE);

  sensor_data.fan[0] = lm78_read_value( LM78_FAN1 );
  sensor_data.fan[1] = lm78_read_value( LM78_FAN2 );
  sensor_data.fan[2] = lm78_read_value( LM78_FAN3 );

  /* TODO: release Spin lock on sensor_data */
}

